from __future__ import annotations

from numbers import Number
from itertools import count

import numpy as np


class Parameters(dict):
    count = count()

    def __init__(self):
        super().__init__()
        self._namespace = {}

    def __str__(self):
        return 'Parameters:\n' + '\n'.join([str(k)+': '+str(v) for k, v in self.items()])

    def __getitem__(self, item):
        if isinstance(item, (list, tuple, np.ndarray)):
            values = []
            for item_i in item:
                values.append(super().__getitem__(item_i))
            return values
        else:
            return super().__getitem__(item)

    @staticmethod
    def _prep_bounds(val: list, p_len: int) -> list:
        # helper function to ensure that bounds and variable are of parameter shape
        if isinstance(val, (Number, bool)) or val is None:
            return [val] * p_len

        elif len(val) == p_len:
            return val

        elif len(val) == 1:
            return [val[0]] * p_len

        else:
            raise ValueError(f'Input {val} has wrong dimensions')

    def add_parameter(self, param: float | list[float], var=None, lb=None, ub=None, names=None) -> list:
        if isinstance(param, (float, int)):
            param = [param]

        p_len = len(param)

        # make list if only single value is given
        var = self._prep_bounds(var, p_len)
        lb = self._prep_bounds(lb, p_len)
        ub = self._prep_bounds(ub, p_len)

        new_keys = []
        for i in range(p_len):
            new_idx = next(self.count)
            if names is not None:
                key = names[i]
            else:
                key = f'_p{new_idx}'
            new_keys.append(key)
            self[key] = Parameter(param[i], var=var[i], lb=lb[i], ub=ub[i])

        return new_keys

    def add(self, name: str | Parameter, value: str | float | int = None, *, var=True, lb=-np.inf, ub=np.inf):
        if isinstance(name, Parameter):
            par = name
        else:
            par = Parameter(name=name, value=value, var=var, lb=lb, ub=ub)
        self[name] = par
        par.eval_allowed = False
        self._namespace[name] = par.value
        par.namespace = self._namespace
        par.eval_allowed = True

    def copy(self) -> Parameters:
        print('huhuuhuh')
        new_para_dict = Parameters()
        for k, v in self.items():
            new_para = v.copy()
            new_para_dict.add(new_para)

        # if len(p) == 0:
        #     return p

        # max_k = int(max(p.keys(), key=lambda x: int(k[2:]))[2:])
        # c = next(p.count)
        # while c < max_k:
        #     c = next(p.count)

        return new_para_dict

    def get_state(self):
        return {k: v.get_state() for k, v in self.items()}


class Parameter:
    """
    Container for one parameter
    """

    def __init__(self, name: str, value: float | str, var: bool = True, lb: float = -np.inf, ub: float = np.inf):
        self.scale = 1
        self.var = bool(var) if var is not None else True
        self.error = None if self.var is False else 0.0
        self.name = name
        self.function = ''
        self.lb = lb if lb is not None else -np.inf
        self.ub = ub if ub is not None else np.inf
        self._value = None
        self.namespace = None
        self.eval_allowed = True
        self._expr = None

        if isinstance(value, str):
            self._expr = value
            self.var = False

        else:
            if self.lb <= value <= self.ub:
                self._value = value
            else:
                raise ValueError('Value of parameter is outside bounds')

            self.init_val = value

            with np.errstate(divide='ignore'):
                # throws RuntimeWarning for zeros
                self.scale = 10**(np.floor(np.log10(np.abs(self.value))))

            if self.scale == 0:
                self.scale = 1.

    def __str__(self) -> str:
        start = ''
        if self.name:
            if self.function:
                start = f'{self.name} ({self.function}):  '
            else:
                start = self.name + ':  '

        if self.var:
            return start + f'{self.value:.4g} +/- {self.error:.4g}, init={self.init_val}'
        else:
            ret_string = start + f'{self.value:}'
            if self._expr is None:
                ret_string += ' (fixed)'
            else:
                ret_string += f' (calc: {self._expr})'

            return ret_string

    def __add__(self, other: Parameter | float | int) -> float:
        if isinstance(other, (float, int)):
            return self.value + other
        elif isinstance(other, Parameter):
            return self.value + other.value

    def __radd__(self, other: Parameter | float) -> float:
        return self.__add__(other)

    @property
    def scaled_value(self):
        return self.value / self.scale

    @scaled_value.setter
    def scaled_value(self, value: float):
        self._value = value * self.scale

    @property
    def value(self) -> float:
        if self._expr is not None and self.eval_allowed:
            return eval(self._expr, self.namespace)
        elif self._value is not None:
            return self._value

    @property
    def scaled_error(self):
        if self.error is None:
            return self.error
        else:
            return self.error / self.scale

    @scaled_error.setter
    def scaled_error(self, value):
        self.error = value * self.scale

    def get_state(self):

        return {slot: getattr(self, slot) for slot in self.__slots__}

    @staticmethod
    def set_state(state: dict):
        par = Parameter(state.pop('value'))
        for k, v in state.items():
            setattr(par, k, v)

        return par

    @property
    def full_name(self):
        name = self.name
        if self.function:
            name += ' (' + self.function + ')'

        return name

    def copy(self) -> Parameter:
        if self._expr:
            val = self._expr
        else:
            val = self._value
        para_copy = Parameter(name=self.name, value=val, var=self.var, lb=self.lb, ub=self.ub)

        return para_copy
